import sys, os, time
import dbus, gobject

gobject.threads_init()

# Set any of these to None to turn that logging off
DEBUG = sys.stderr
ERROR = sys.stderr
WARN = sys.stderr

TEST_STRING = "xesam"

def get_time () : return time.strftime("%Y-%m-%d %H:%M:%S")

# If we are not in debug mode xesam.debug(msg) just resolves to a pass
if DEBUG:
	def debug (msg):
		if DEBUG : print >> DEBUG, "[DEBUG, %s] %s" % (get_time(), msg)
else:
	def debug (msg): pass

# We always log warnings (as opposed to debug)
def warn (msg):
	if WARN : print >> WARN, "[WARN, %s] %s" % (get_time(), msg)

# We always log errors (as opposed to debug)	
def error (msg):
	if ERROR : print >> ERROR, "[ERROR, %s] %s" % (get_time(), msg)
	
HELPFILE = os.path.join(os.path.dirname(__file__),"help.txt")

MAX_UINT = 4294967295

# These are the environment variables to set dbus service parameters
APP_NAME = "XESAM_APPLICATION"
OBJ_NAME = "XESAM_OBJECT"
IFACE = "XESAM_INTERFACE"

# Namespaces
NAMESPACE		= "http://freedesktop.org/standards/xesam/1.0/"
NAMESPACE_QUERY	= NAMESPACE + "query"
NAMESPACE_CORE 	= NAMESPACE + "core#"

# Set up a dict with dbus service info
# Use default values if environment variables are not set
SEARCH_SERVICE = \
{
	"application" 	: not os.getenv(APP_NAME) and "org.freedesktop.xesam.searcher" or os.getenv(APP_NAME),
	"object"		: not os.getenv(OBJ_NAME) and "/org/freedesktop/xesam/searcher/main" or os.getenv(OBJ_NAME),
	"interface" 	: not os.getenv(IFACE) and "org.freedesktop.xesam.Search" or os.getenv(IFACE)
}

class Property:
	"""
	A genric representation of a Session property
	"""
	def __init__ (self, name, value, string_parser):
		"""
		@param string_parser: a function that returns a proper object parsed
		    from a input string. Fx convert comma separated list to real list.
		"""
		self._name = name
		self._value = value# Session Properties
		self._string_parser = string_parser
	
	def get_name (self):
		return self._name
	
	def get_value(self):
		return self._value
	
	def set_value (self, value):
		self._value
	
	def set_value_from_string (self, value):
		self._value = self._string_parser(value)
	
	def __str__ (self):
		return self._name + "=" + self._value


def uint (value):
	"""
	Helper method to wrap integers in dbus UInt32.
	"""
	if type (value) == dbus.UInt32:
		return value
	else :
		return dbus.UInt32 (value)

def pyvalue (value, dummy=False):
	"""
	Convert a dbus variant type to a Python type.
	If dummy=True this method will just return the
	passed paramter as is.
	"""
	if dummy:
		return value
	
	if type(value) == dbus.Array:
		# Do deep conversions of arrays
		value = list(value[0:len(value)])
		if len (value) == 0:
			return []
		for i in range(len(value)):
			value[i] = pyvalue(value[i])
	elif type(value) == dbus.String:
		value = unicode(value)
	elif type(value) == dbus.Boolean:
		value = bool(value)
	elif type(value) == dbus.String:
		value = unicode(value)
	elif type(value) == dbus.Int32:
		value = int(value)
	elif type(value) == dbus.UInt32:
		value = int(value)
	return value

def prop_val (prop, val):
	"""
	Given two strings prop and val, return val in a format that is specified
	by the Xesam session properties. Values that should be converted to lists
	should be comma separated strings (without spaces).
	"""
	if prop.startswith ("hit.fields") or prop.startswith("sort.fields") or prop.startswith("sort.order"):
		# this also catches hit.fields.extended
		return val.split(",")
	elif prop == "vendor.search.live":
		return bool(val)
	elif prop == "hit.snippet.length" or \
	     prop == "vendor.version" or \
	     prop == "vendor.xesam" or \
	     prop == "vendor.maxhits":
		return uint(val)
	else:
		raise NameError ("Unknown property %s for value %s" % (prop,val))	
		
# Session Properties
SESSION_HIT_FIELDS 					= "hit.fields"
SESSION_HIT_FIELDS_EXTENDED			= "hit.fields.extended"
SESSION_HIT_SNIPPET_LENGTH			= "hit.snippet.length"
SESSION_SORT_FIELDS 				= "sort.fields"
SESSION_SORT_ORDER					= "sort.order"
SESSION_VENDOR_ID					= "vendor.id"
SESSION_VENDOR_VERSION				= "vendor.version"
SESSION_VENDOR_DISPLAY				= "vendor.display"
SESSION_XESAM						= "vendor.xesam"
SESSION_VENDOR_ONTOLOGY_FIELDS		= "vendor.ontology.fields"
SESSION_VENDOR_ONTOLOGY_CONTENTS	= "vendor.ontology.contents"
SESSION_VENDOR_ONTOLOGY_STORAGES	= "vendor.ontology.sources"
SESSION_VENDOR_SEARCH_LIVE 			= "vendor.search.live"
SESSION_VENDOR_EXTENSIONS			= "vendor.extensions"
SESSION_VENDOR_MAXHITS				= "vendor.maxhits"
SESSION_VENDOR_SORT_MAXFIELDS		= "vendor.sort.maxfields"
SESSION_VENDOR_SORT_FIELDS			= "vendor.sort.fields"

# Session properties and their default values
# The FIXMEs in the default values are due to the fact the Python
# dbus bindings doesn't know how to serialize empty lists
PROPERTIES = \
{	
	"hit.fields"			    : Property ("hit.fields", ["xesam:url"], lambda x : x.split(",")),
	"hit.fields.extended"		: Property ("hit.fields.extended", dbus.Array(signature="s"), lambda x : x.split(",")),
	"hit.snippet.length"		: Property ("hit.snippet.length", uint(200), uint),
	"sort.fields"			    : Property ("sort.fields", ["xesam:relevancyRating"], lambda x : x),
	"sort.order"			    : Property ("sort.order", dbus.Array(signature="s"), lambda x : x.split(",")),
	"vendor.id"			        : Property ("vendor.id", "Unknown", str),
	"vendor.version"		    : Property ("vendor.version", uint(90), uint),
	"vendor.display"		    : Property ("vendor.display", "Unknown Vendor", str),
	"vendor.xesam"			    : Property ("vendor.xesam", uint(100), uint),
	"vendor.ontology.fields"	: Property ("vendor.ontology.fields", ["xesam:url"], lambda x : x),
	"vendor.ontology.contents"	: Property ("vendor.ontology.contents", ["xesam:Content"], lambda x : x),
	"vendor.ontology.sources"	: Property ("vendor.ontology.storages", ["xesam:Source"], lambda x : x),
	"vendor.search.live" 		: Property ("vendor.search.live", True, bool),
	"vendor.extensions"		    : Property ("vendor.extensions", dbus.Array(signature="s"), lambda x : x.split(",")),
	"vendor.maxhits"		    : Property ("vendor.maxhits", uint(10000), lambda x : FIXME),
	"vendor.sort.maxfields"	    : Property ("vendor.sort.maxfields", uint(10000), lambda x : FIXME),
	"vendor.sort.fields"	    : Property ("vendor.sort.fields", dbus.Array(signature="s"), lambda x : x.split(",")),
}


#
# States as reported by GetState and StateChanged
#
STATE_IDLE = "IDLE"
STATE_UPDATE = "UPDATE"
STATE_FULL_INDEX = "FULL_INDEX"
STATES = [STATE_IDLE, STATE_UPDATE, STATE_FULL_INDEX]

