import xesam
import xesam.query

import dbus
import dbus.glib
import gobject
from gobject import GObject
from cgi import escape

USER_QUERY_TEMPLATE = \
"""
<request>
  <userQuery>
    %s
  </userQuery>
</request>
"""

class UnsupportedPropertyValue (Exception):
	def __init__(self, msg):
		Exception.__init__(self, msg)

class Session (GObject):
	"""
	A proxy object for a Session
	"""
	
	__gsignals__ = {
		# Emitted when the connection to the search engine has been established
		"ready" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
		"closed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
	}
	
	__gproperties__ = {
		"hit.fields" : (gobject.TYPE_PYOBJECT, "The properties to return in GetHits",
						"Fields to return in the GetHits method. A list of field names. Defaults to ['xesam:url']", 
						gobject.PARAM_READWRITE),
		"hit.fields.extended" : (gobject.TYPE_PYOBJECT, "Extra properties that are likely to be requested",
						"Extra fields that are likely to be requested for each hit. This is only a hint to the search engine", 
						gobject.PARAM_READWRITE),
		"hit.snippet.length" : (gobject.TYPE_UINT, "Requested length in chars of hit excerpt",
						"A hint to the search engine as to what length in chararcters the hit content excerpts should have",
						1, xesam.MAX_UINT, 200, gobject.PARAM_READWRITE),
		"sort.fields" : (gobject.TYPE_PYOBJECT, "List of fields to sort after",
						"List of fields to sort after in the given order. Defaults to ['xesam:relevancyRating']",
						gobject.PARAM_READWRITE),
		"sort.order" : (gobject.TYPE_PYOBJECT, "Order of hit sorting per field",
						"List of orderings, 'ascending' or 'descending', for the fields specified in sort.fields",
						gobject.PARAM_READWRITE),
		"vendor.id" : (gobject.TYPE_STRING, "Unique id identifying the search engine provider",
						"Non-versioned id uniquely identifying the search engine provider",
						"Unknown id", gobject.PARAM_READWRITE),
		"vendor.version" : (gobject.TYPE_UINT, "An integer describing the search engine version",
						"An integer describing the search engine version",
						0, xesam.MAX_UINT, 1, gobject.PARAM_READWRITE),
		"vendor.display" : (gobject.TYPE_STRING, "A string suitable for display that describes the search engine",
						"A string suitable for display that describes the search engine",
						"Unknown search engine", gobject.PARAM_READWRITE),
		"vendor.xesam" : (gobject.TYPE_UINT, "Version of implemented Xesam search spec",
						"The version of the Xesam search spec that the vendor implements",
						1, xesam.MAX_UINT, 1, gobject.PARAM_READWRITE),
		"vendor.ontology.fields" : (gobject.TYPE_PYOBJECT, "The metadata fields supported by the vendor",
						"The metadata fields supported/indexed by the search engine",
						gobject.PARAM_READWRITE),
		"vendor.ontology.contents" : (gobject.TYPE_PYOBJECT, "The known content types",
						"The content types supported by the search engine",
						gobject.PARAM_READWRITE),
		"vendor.ontology.sources" : (gobject.TYPE_PYOBJECT, "The known source types",
						"The source types understood by the search engine",
						gobject.PARAM_READWRITE),
		"vendor.search.live" : (gobject.TYPE_BOOLEAN, "Maps to the vendor.search.live property in the serverside session",
						"Whether or not the search-invalidated signal will be used",
						True, gobject.PARAM_READWRITE),
		"vendor.extensions" : (gobject.TYPE_PYOBJECT, "Supported query extensions",
						"A list of supported extensions to the Xesam query language",
						gobject.PARAM_READWRITE),
		"vendor.maxhits" : (gobject.TYPE_UINT, "Maximum number of hits",
						"Maximum number of hits a client is able to retrieve",
						1, xesam.MAX_UINT, 10000, gobject.PARAM_READWRITE),
		"vendor.sort.fields" : (gobject.TYPE_PYOBJECT, "The fields that the server can sort on",
						"The server is only able to sort on the listed fields, empty means all fields. The default is the empty list",
						gobject.PARAM_READWRITE),
		"vendor.sort.maxfields" : (gobject.TYPE_UINT, "Maximum number of fields server can sort by",
						"Maximum number of fields that can be sorted on simultaneously. Zero means unlimited",
						0, xesam.MAX_UINT, 10000, gobject.PARAM_READWRITE),
	}
	
	
	def __init__ (self, use_dbus_types=False):
		GObject.__init__ (self)
		
		# Set up dbus connection to the Xesam Service
		self._bus = dbus.SessionBus()
		self._proxy_obj = self._bus.get_object(xesam.SEARCH_SERVICE["application"], xesam.SEARCH_SERVICE["object"])
		self._xesam = dbus.Interface(self._proxy_obj, xesam.SEARCH_SERVICE["interface"])
		
		self._use_dbus_types = use_dbus_types
		
		self._closed = False
		self._handle = xesam.pyvalue(self._xesam.NewSession(), dummy=self._use_dbus_types)
	
	def get_searcher (self):
		"""
		Return the dbus.Interface used for this Session
		"""
		return self._xesam
	
	def close (self):
		if self._closed :
			return
		self._xesam.CloseSession (self._handle) # FIXME blocking dbus call
		self._closed = True
		self.emit ("closed")
		
	def get_handle (self):
		"""
		Get the session handle as registered on the search engine
		"""
		return self._handle
	
	def do_get_property(self, prop):
		# This trick allows us to call this methods directly for debugging purposes
		if not isinstance(prop, basestring):
			name = prop.name.replace("-", ".")
		else:
			name = prop
		val = self._xesam.GetProperty (self._handle, name)
		val = xesam.pyvalue(val, dummy=self._use_dbus_types)
		return val

	def do_set_property(self, prop, value):
		# This trick allows us to call this methods directly for debugging purposes
		if not isinstance(prop, basestring):
			name = prop.name.replace("-", ".")
		else:
			name = prop
		used_value = self._xesam.SetProperty (self._handle, name, value)
		if not used_value == value:
			raise UnsupportedPropertyValue(name + "=" + value)
		

class Search (GObject) :
	"""
	Proxy object representing a server side search
	"""
	
	__gsignals__ = {
		"hits-added" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_INT]),
		"invalidated" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_STRING]),
		"done" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_UINT, gobject.TYPE_UINT, gobject.TYPE_STRING]),
		"closed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, []),
	}
	
	def __init__ (self, session, query):
		GObject.__init__ (self)
		if isinstance (query, basestring):
			self._query = query
		elif isinstance (query, xesam.query.Query):
			self._query = query.xml()
		else:
			raise QueryException ("query argument must a string type or xesam.query.Query instance")
		
		self._invalid = False
			
		self._hit_props = session.get_property ("hit.fields")
		self._last_hit_id = 0
		
		self._searcher = session.get_searcher()
		
		# Start the search
		self._handle = str(self._searcher.NewSearch(session.get_handle(), self._query)) # FIXME: This is a sync dbus call, shame on me
		
		# Indicates whether start() has been called
		self._started  = False
		
		# Connect to signals with appropriate match rules for our search handle
		self._searcher.connect_to_signal ("HitsAdded", self._on_hits_added, arg0=self._handle)
		self._searcher.connect_to_signal ("SearchInvalidated", self._on_search_invalidated, arg0=self._handle)
		self._searcher.connect_to_signal ("SearchDone", self._on_search_done, arg0=self._handle)
				
	def _on_hits_added (self, handle, count):
		self.emit ("hits-added", count)
	
	def _on_search_invalidated (self, handle, debug_msg):
		self._invalid = True
		self.emit ("invalidated", debug_msg)
	
	def _on_search_done (self, handle, hit_count, exit_state, msg):
		self.emit ("done", hit_count, exit_state, msg)
	
	def start(self):
		if self._invalid:
			raise Exception("Can no start invalidated search")		
		self._searcher.StartSearch(self._handle) # FIXME: This is a sync dbus call, shame on me
		self._started = True
	
	def get_handle (self):
		return self._handle
	
	def get_hit_count (self):
		if self._invalid:
			xesam.error("get_hit_count() on invalidated search")
		if not self._started:
			xesam.error ("get_hit_count() on unstarted seearch")
		return xesam.pyvalue(self._searcher.GetHitCount (self._handle))
	
	def get_hits (self, count):
		if self._invalid:
			xesam.error("get_hits() on invalidated search")
		if not self._started:
			xesam.error ("get_hits() on unstarted search")
			
		result = []
		hits = self._searcher.GetHits(self._handle, count)
		
		for hit_data in hits:
			result.append(Hit(self._last_hit_id, self._hit_props, hit_data))	
			self._last_hit_id += 1
		
		return result
	
	def get_hit_data (self, hits, hit_props):
		"""
		Get additional (or renewed) metadata for a list of Hits.
		The returned list of Hit objects have the same ids (in the same order)
		as the ones provided in the arguments.
		"""
		if self._invalid:
			xesam.error("get_hit_data() on invalidated search")
		if not self._started:
			xesam.error ("get_hit_data on unstarted seearch")
			
		hit_ids = [hit.get_id() for hit in hits]
		return self.get_hit_data_from_ids(hit_ids, hit_props)
	
	def get_hit_data_from_ids (self, hit_ids, hit_props):
		"""
		Get additional (or renewed) metadata from a set of hits
		given their ids.
		"""
		data_set = self._searcher.GetHitData(self._handle, hit_ids, hit_props)
		result = []
		for i in range(len(data_set)):
			result.append (Hit(hit_ids[i], hit_props, data_set[i]))
		
		return result
	
	def close (self):
		if self._invalid:
			xesam.error("close() on invalidated search")
		if not self._started:
			xesam.debug ("Warning: Closing unstarted search.")
		self._searcher.CloseSearch (self._handle)
		self.emit ("closed")
	
	def is_invalid(self):
		return self._invalid

class UserSearch (Search):
	def __init__ (self, session, user_search):
		Search.__init__ (self, session, USER_QUERY_TEMPLATE % escape(user_search))

class Hit :
	"""
	You can look up metadata values as in a dict by looking up
	directly on this object. Like 
	
		uri = my_hit["uri"]
		
	"""
	def __init__ (self, hit_id, fields, values):
		self._id = hit_id
		self._data = {}
		
		self._data["id"] = hit_id
		
		assert len(fields) == len(values)
		
		# Map property names to their corresponding values
		for i in range(len(fields)):
			self._data[fields[i]] = xesam.pyvalue(values[i])
		
	def __getitem__ (self, key):
		return self._data[key]
	
	def __iter__ (self):
		return self._data.iteritems()
	
	def get_id (self):
		return self._id
	
	def get_data (self):
		return self._data
	
	def get_field_count (self):
		return len (self._data)
	
	def __str__ (self):
		result = "[Hit:%s]" % self.get_id()
		for field,val in self._data.iteritems():
			result += "\n%s=%s" % (field,val)
		return result


class EngineState (GObject):
	"""
	And object that monitors the search engine's state and dynamically
	updates when the state of the engine changes.
	"""
	
	__gsignals__ = {
		"changed" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
	}
	
	def __init__ (self, xesam_dbus=None):
		"""
		Create a new engine state monitor with the given dbus connection
		to the engine. If xesam_dbus is None a new connection will be made.
		"""
		GObject.__init__ (self)
		if xesam_dbus is None:
			_bus = dbus.SessionBus()
			_proxy_obj = _bus.get_object(xesam.SEARCH_SERVICE["application"], xesam.SEARCH_SERVICE["object"])
			xesam_dbus = dbus.Interface(_proxy_obj, xesam.SEARCH_SERVICE["interface"])
			
		self._xesam = xesam_dbus
		self._state_name = None
		self._state_value = None
		
		self._xesam.connect_to_signal ("StateChanged", self._on_state_changed)
		self._xesam.GetState(reply_handler=lambda state_info : self._on_state_changed(state_info), error_handler=self._receive_error)
	
	def get_state (self):
		"""
		The state in which the engine is. This can be "IDLE", "UPDATE", or
		"FULL_INDEX". 
		"""
		self._ensure_state_info()		
		return self._state_name
	
	def get_value (self):
		"""
		The value associated with the engine state. If the engine state is IDLE
		this value is undefined. If the state is UPDATE or FULL_INDEX this will
		be a percentage number of how much of the task is completed.
		
		If the engine state has not yet been fetched this method will return
		["UNKNOWN", "0"]
		"""
		self._ensure_state_info()
		return self._state_value
		
	def _ensure_state_info (self):
		if self._state_name is None or self._state_value is None:
			self._state_name = "UNKNOWN"
			self._state_value = "0"
	
	def _on_state_changed (self, state_info):
		self._state_name = xesam.pyvalue(state_info[0])
		self._state_value = xesam.pyvalue(state_info[1])
		self.emit ("changed", [self._state_name, self._state_value])
	
	def _receive_error (self, error):
		xesam.error ("Error reading search engine state: %s" % error)
		
if gobject.pygobject_version < (2,8,0):	
	gobject.type_register (Session)
	gobject.type_register (Search)
	gobject.type_register (EngineState)
				
