"""
This module contains classes for xesam query construction
"""

import xesam
import xml.sax
from xml.sax.saxutils import escape, prepare_input_source
import os

SELECTORS=["equals", "contains", "lessThan", "lessThanEquals", "greaterThan", 
		   "greaterThanEquals", "startsWith", "fullText", "inSet", "regExp", 
		   "proximity", "category"]

COLLECTORS = ["and", "or"]


class QueryException (Exception):
	def __init__ (self, msg=None):
		if msg is None:
			Exception.__init__ (self)
		else:
			Exception.__init__ (self, msg)

class QueryTypeMixupException (QueryException):
	"""
	This exception is thrown if you try to change a standard Query into a user
	query or vice versa. Fx. by calling set_clause and then set_user_string.
	"""
	def __init__ (self, msg=None):
		QueryException.__init__ (self, msg)

class InvalidValueException (QueryException):
	"""
	Thrown if you try to set a value on a Value that does not map to
	its supported type. Fx if you try to set the value "xyz" on a 
	BooleanValue.
	"""
	def __init__ (self, msg=None):
		QueryException.__init__ (self, msg)

class XmlSerializable:
	"""
	And interface for objects serializable to xml fragments. Any fragment should
	become valid xml if the xml header is added.
	"""
	def __init__ (self):
		pass
	
	def xml (indent=0):
		"""
		Return an xml fragment representation of the object. If indent is provided
		the fragment will be indented this many spaces.
		"""
		raise NotImplementedError
	
	def set_attributes (self, attributes):
		"""
		Takes a dict and adds them to the attributes
		of the top level element
		"""
		raise NotImplementedError
	
	def get_attributes (self):
		"""
		Return a dict of all attribute-value pairs of the 
		top level element
		"""
		raise NotImplementedError
		
	def get_name (self):
		"""
		Return the tag name of the topmost element
		"""
		raise NotImplementedError

class Query (XmlSerializable):
	"""
	Abstract class that represents a generic query.
	"""
	def __init__ (self):
		XmlSerializable.__init__ (self)

class UserQuery (XmlSerializable, Query):
	"""
	This clause is used to embed a user search string as described in http://wiki.freedesktop.org/wiki/XesamUserSearchLanguage.
	
	Represents a query string as entered by the user. This clause must be the
	only clause in the Query object.
	"""
	def __init__ (self, user_search=""):
		Query.__init__ (self)
		self._string = user_search
		
	def get_boost (self):
		return 1.0
	
	def get_string (self):
		return self._string
	
	def set_string (self, user_search):
		"""
		Set the user search string. Syntax is described on 
		http://wiki.freedesktop.org/wiki/XesamUserSearchLanguage
		"""
		self._string = user_search
	
	def set_attributes (self, attributes):
		pass
	
	def get_attributes (self):
		return {}
	
	def get_name (self):
		return "userQuery"
		
	def xml (self, indent=0):
		return \
		'<?xml version="1.0" encoding="UTF-8"?>\n' + \
		indent*" " + '<request xmlns="' + xesam.NAMESPACE_QUERY + '">\n' + \
		(indent+2)*" " + "<userQuery>\n"   + \
		(indent+4)*" " + escape(self._string) + "\n" + \
		(indent+2)*" " + "</userQuery>\n" +\
		indent*" " + "</request>" 

	
class CompositeQuery (Query):
	def __init__ (self):
		Query.__init__(self)
		self._clause = None
		self._attributes = {}
		self._content = None
		self._source = None
        
	def set_clause (self, clause):
		"""
		Set the root clause of this query object.
		You should only call either set_user_string or set_clause on any single Query object.
		"""
		self._clause = clause
	
	def get_clause (self):
		return self._clause
	
	def get_contents (self):
		"""
		Return the list of content types to search. An empty list indicates
		that all reasonable content types will be searched. 
		"""
		if self._attributes.has_key("content"):
			return map (lambda x : x.strip(), self._attributes["content"].split(","))
		else:
			return []
	
	def get_source (self):
		"""
		Return the list of storage media searched. An empty list indicates
		that all reasonable media will be searched. 
		"""
		if self._attributes.has_key("source"):
			return map (lambda x : x.strip(), self._attributes["source"].split(","))
		else:
			return []
	
	def set_content (self, content_list):
		"""
		Accepts a list of content types to be searched
		"""
		if type(content_list) == list:
			self._attributes["content"] = ", ".join(content_list)
		elif isinstance (content_list, basestring):
			self._attributes["content"] = content_list
		else:
			raise TypeError ("Invalid content attribute format %s" % type(content_list))
	
	def set_source (self, source_list):
		"""
		Accepts a list of strings defining what source/storage types should
		be searched
		"""
		if type(source_list) == list:
			self._attributes["source"] = ", ".join(source_list)
		elif isinstance (source_list, basestring):
			self._attributes["source"] = source_list
		else:
			raise TypeError ("Invalid content attribute format %s" % type(content_list))
	
	def set_attributes (self, attributes):
		if attributes.has_key("content"):
			self.set_content(attributes["content"])
		if attributes.has_key ("source"):
			self.set_source(attributes["source"])
	
	def get_attributes (self):
		return self._attributes
	
	def get_name (self):
		return "query"
	
	def xml (self, indent=0):
		if self._clause is None:
			raise QueryException ("Query has no Clause")
		
		return \
		'<?xml version="1.0" encoding="UTF-8"?>\n' + \
		indent*" " + '<request xmlns="' + xesam.NAMESPACE_QUERY + '">\n' + \
		indent*" " + "  <query" + self._attribute_string() + ">" + "\n" + \
		indent*" " + self._clause.xml(indent+4) + "\n" +\
		indent*" " + "  </query>\n" + \
		indent*" " + "</request>"
	
	def _attribute_string (self):
		return reduce (lambda x, y : x + ' %s="%s"' % (y[0],y[1]), self._attributes.iteritems(), "")

class Clause (XmlSerializable):
	"""
	A Query that is not described by a user string is described by a series of (possibly nested) clauses.
	"""
	def __init__ (self):
		XmlSerializable.__init__ (self)
		self._attributes = {}
		self._parent = None
	
	def get_name (self):
		"""
		Return the tag name for this query element/clause
		"""
		return self._name
	
	def set_negate (self, boolean):
		"""
		Whether or not to do boolean negation on the matching
		"""
		self._attributes["negate"] = boolean
	
	def is_negated (self):
		if not self._attributes["negate"]:
			return False
		return bool(self._attributes["negate"])
	
	def set_boost (self, boost):
		self._attributes["boost"] = float(boost)
	
	def set_attributes (self, attributes):
		if attributes.has_key("boost"):
			self._attributes["boost"] = float(attributes["boost"])
		if attributes.has_key("negate"):
			self._attributes["negate"] = bool(attributes["negate"])
	
	def get_attributes (self):
		return self._attributes
			
	def get_boost (self):
		if not self._attributes["boost"]:
			return 1.0
		return float(self._attributes["boost"])
	
	def set_parent (self, parent_clause):
		self._parent = parent_clause
	
	def get_parent (self):
		return self._parent
	
	def get_top_parent (self):
		parent = self
		while parent.get_parent():
			parent = parent.get_parent()
		return parent
	
	def _attribute_string (self):
		return reduce (lambda x, y : x + ' %s="%s"' % (y[0],y[1]), self._attributes.iteritems(), "")
	
class CollectorClause (Clause):
	"""
	A collector clause is a clause that collects the results of other clauses
	and does matching based on these results. The prime examples are boolean
	AND and OR.
	"""
	def __init__ (self, name="and"):
		Clause.__init__ (self)
		self._children = []
		self._name = name
		
	def append_clause (self, clause):
		clause.set_parent (self)
		self._children.append (clause)
	
	def append_clauses (self, clauses):
		for clause in clauses:
			clause.set_parent (self)
		self._children.extend(clauses)
	
	def get_children (self):
		return self._children
	
	def clear_children (self):
		for child in self._children:
			child.set_parent (None)
		self._children = []
	
	def xml (self, indent=0):
		return \
		indent*" " + "<%s%s>\n" % (self._name, self._attribute_string()) + \
		self._inner_xml (indent+2) + "\n" + \
		indent*" " + "</%s>" % self._name
		
		
	def get_name (self):
		return self._name	
	
	def _inner_xml (self, indent=0):
		inner = ""
		for clause in self._children:
			inner += clause.xml(indent) + "\n"
		return inner[:-1]
			
	
class SelectorClause (Clause):
	
	def __init__ (self, name, check=True):
		if not name in SELECTORS and check:
			raise QueryException ("Unknown selector '%s'" % name)
		
		Clause.__init__ (self)
		self._name = name
		self._fields = []
		self._value = None
		
	def append_field (self, field):
		if not isinstance (field, Field):
			raise QueryException ("non Field object passed to append_field, got a " + str(field.__class__))
		self._fields.append (field)
	
	def get_fields (self):
		return self._fields
	
	def set_value (self, value):
		if not isinstance (value, Value):
			raise QueryException ("non Value object passed to set_value, got a " + str(value.__class__))
		self._value = value
	
	def get_value (self):
		return self._value
	
	def xml (self, indent=0):
		if len(self._fields) == 0:
			raise QueryException ("%s defines no fields" % self.__class__)
		if self._value is None:
			raise QueryException ("%s defines no value" % self.__class__)
			
		return \
		indent*" " + "<%s%s>\n" % (self._name, self._attribute_string()) + \
		reduce (lambda x, y : x + y.xml(indent+2) + "\n", self._fields, "") + \
		self._value.xml(indent+2) + "\n" + \
		indent*" " + "</%s>" % self._name

class MultiValueClause (SelectorClause):

	def __init__(self, name):
		SelectorClause.__init__(self, name)
		self._values = []
		
	def append_value (self, value):
		"""
		Append a value to the set of values that are matched
		"""
		if not isinstance (value, Value):
			raise QueryException ("Trying to append non Value as a value, got: %s" % type(value))
			
		self._values.append(value)
		
	def set_value (self, value):
		raise QueryException (str(self.__class__) + ".set_value is ambigous, use append_value instead")
	
	def xml (self, indent=0):
		if len(self._fields) == 0:
			raise QueryException (str(self.__class__) + " has no fields")
		if len(self._values) == 0:
			raise QueryException (str(self.__class__) + " has no Values")
			
		return \
		indent*" " + "<%s%s>\n" % (self._name, self._attribute_string()) + \
		reduce (lambda x, y : x + y.xml(indent+2) + "\n", self._fields, "") + \
		reduce (lambda x, y : x + y.xml(indent+2) + "\n", self._values, "") + \
		indent*" " + "</%s>" % self._name


class EqualsClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "equals")

class ContainsClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "contains")

class LessThanClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "lessThan")

class LessThanEqualsClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "lessThanEquals")

class GreaterThanClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "greaterThan")

class GreaterThanEqualsClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "greaterThanEquals")

class StartsWithClause (SelectorClause):
	def __init__ (self):
		SelectorClause.__init__ (self, "startsWith")

class InSetClause (MultiValueClause):
	def __init__ (self):
		MultiValueClause.__init__ (self, "inSet")

class FullTextClause (SelectorClause): 
	def __init__(self):
		SelectorClause.__init__(self,"fullText")
	
	def append_field (self, field):
		raise QueryException ("Cannot assign fields to a FullTextClause. Got field %s" % field)
	
	def xml (self, indent=0):
		return \
		indent*" " + "<%s%s>\n" % (self._name, self._attribute_string()) + \
		self._value.xml(indent+2) + "\n" + \
		indent*" " + "</%s>" % self._name

class ProximityClause (MultiValueClause): 
	def __init__(self):
		MultiValueClause.__init__(self, "proximity")
		self._distance = 10
	
	# override parent method
	def append_field (self, field):
		if not (isinstance (field, Field) or isinstance (field, FullTextFields)):
			raise QueryException ("non Field or FullTextFields object passed to append_field, got a " + str(field.__class__))
		if len (self._fields) > 0 and isinstance (field, FullTextFields):
			raise QueryException ("FullTextFields must be the sole Field declaration")
		#FIXME Make sure we *only* have a FullTextFields or one or more Fields
		self._fields.append (field)
		
	def set_attributes (self, attributes):
		if attributes.has_key("distance"):
			self._distance = attributes["distance"]
		
		# Chain up to super class' set_attributes
		SelectorClause.set_attributes (self, attributes)
	
	def set_distance (self, value):
		self._distance = int(value)
	
	def get_distance (self):
		return self._distance
	
	def _attribute_string (self):
		# Don't set the distance attribute if it is the default value 10
		return MultiValueClause._attribute_string(self) +  (self._distance != 10 and (' distance="%s"' % self._distance) or "")

class RegExpClause (SelectorClause): 
	def __init__(self):
		SelectorClause.__init__(self,"regExp")
	
	# override parent method
	def append_field (self, field):
		if not (isinstance (field, Field) or isinstance (field, FullTextFields)):
			raise QueryException ("non Field or FullTextFields object passed to append_field, got a " + str(field.__class__))
		if len (self._fields) > 0 and isinstance (field, FullTextFields):
			raise QueryException ("FullTextFields must be the sole Field declaration")
		#FIXME Make sure we *only* have a FullTextFields or one or more Fields
		self._fields.append (field)

class CategoryClause (SelectorClause):
	def __init__(self):
		SelectorClause.__init__(self,"category")
	
	# override parent method
	def append_field (self, field):
		raise QueryException ("category selector does not match on fields")
	
	def set_content (self, category):
		"""
		Set what content type to match. Fx. xesam:Audio
		"""
		self.attributes["content"] = category
	
	def set_source(self, category_name):
		"""
		Set the source type to match. Fx. xesam:File.
		"""
		self.attributes["source"] = category_name
	
	def set_attributes (self, attrs):
		if attrs.has_key("content"):
			self._attributes["content"] = attrs["content"]
		if attrs.has_key("source"):
			self._attributes["source"] = attrs["source"]
		
		SelectorClause.set_attributes(self, attrs)
	
	def xml (self, indent=0):
		return \
		indent*" " + "<category" + self._attribute_string() + "/>" # MISSING name value attrs
	
class AndClause (CollectorClause):
	def __init__ (self):
		CollectorClause.__init__ (self, "and")

class OrClause (CollectorClause):
	def __init__ (self):
		CollectorClause.__init__ (self, "or")

class Field (XmlSerializable):
	def __init__ (self, name=None):
		XmlSerializable.__init__ (self)
		self._field_name = name
	
	def get_name (self):
		return "field"
	
	def get_field_name (self):
		return self._field_name
	
	def set_field_name (self, name):
		self._field_name = name
	
	def set_attributes (self, attrs):
		pass
	
	def xml (self, indent=0):
		return indent*" " + '<field name="%s"/>' % self._field_name

class FullTextFields (XmlSerializable):
	"""
	A special element to use instead of a Field in certain selectors
	that behaves as though the selector queries all the fields a 
	FullTextClause would.
	
	Only allowed for ProximityClause and RegExpClause.
	
	While this is really not a field, it makes sense to treat it as so.
	"""
	def __init__ (self):
		XmlSerializable.__init__ (self)
	
	def get_attributes (self) : return {}
	
	def set_attributes (self, attrs) : pass
	
	def get_name (self) : return "fullTextFields"
	
	def xml (self, indent=0): return " "*indent + "<fullTextFields/>"
		

class Value (XmlSerializable):
	def __init__ (self, value=None, tybe=None, name=None, attributes=None):
		XmlSerializable.__init__ (self)
		self._attributes = (not attributes and [{}] or [attributes])[0]
		self._value = value
		self._type = tybe
		self._name = name
		
		if tybe is not None:
			self._check_type (tybe, value)
	
	def new_by_name (name, value=None):
		"""
		Static factory method to create a Value object with the proper
		class matching a given name.
		"""
		if name == "integer" :
			return value and IntegerValue (value) or IntegerValue ()
		elif name == "boolean" :
			return value and BooleanValue (value) or BooleanValue ()
		elif name == "string" :
			return value and StringValue (value) or StringValue ()
		
		raise QueryException ("Unknown value type: " + name)
		
		
	def set (self, val):
		self._value = self._marshal (val)
	
	def get (self):
		return self._value
	
	def get_name (self):
		return self._name
	
	def set_attribute (self, name, val):
		self._attributes[name] = str(val)
	
	def set_attributes (self, attributes) :
		"""
		Add a dict of attribute->value pairs to the attributes of this value.
		"""
		self._attributes.update (attributes)
	
	def get_attribute (self, name):
		return self._attributes[name]
	
	def get_attributes (self):
		"""
		Return a of all attribute->value pairs on this Value
		"""
		return self._attributes
	
	def _attribute_string (self):
		return reduce (lambda x, y: x + ' %s="%s"' % y, self._attributes.iteritems(), "")
	
	def xml (self, indent=0):
		return indent*" " + "<%s%s>%s</%s>" % (self._name, self._attribute_string(), self._value, self._name)
	
	def _marshal (self, val):
		try:
			if self._type == int:
				return int(val)
			elif self._type == str:
				return escape(str(val))
			elif self._type == bool:
				return bool(val)
			else:
				raise QueryException ("Value object type confusion: " + str(self._type))
		except ValueError:
			raise InvalidValueException ("Error converting '%s' to type %s" % (val, self._type) )
	
	def _check_type (self, tybe, val):
		"""
		Throws a InvalidValueException if val is not of the type tybe
		"""
		if not type(val) == tybe:
			raise InvalidValueException("Expected " + str(tybe) + " but got " + str(type(val)))
	
class IntegerValue (Value):
	def __init__ (self, value=0, attrs=None):
		Value.__init__(self, value, tybe=int, name="integer", attributes=attrs)
		
class DateValue (Value):
	def __init__ (self, value="2005-11-29T05:20:00Z", attrs=None):
		Value.__init__(self, value, tybe=str, name="date", attributes=attrs)

class BooleanValue (Value):
	def __init__ (self, value=False, attrs=None):
		Value.__init__(self, value, tybe=bool, name="boolean", attributes=attrs)

class StringValue (Value):
	def __init__ (self, value="", attrs=None):
		Value.__init__(self, value, tybe=str, name="string", attributes=attrs)
		self._phrase = True
		self._case_sensitive = False
		self._diacritic_sensitive = True
		self._slack = 0
		self._ordered = True
		self._stemming = True
		self._language = "UK_en"
		self._fuzzy = False


CLAUSES = \
{
	"query" 			: lambda : CompositeQuery(),
	"userQuery" 		: lambda : UserQuery (),
	"equals"			: lambda : EqualsClause(), 
	"contains" 			: lambda : ContainsClause (), 
	"lessThan" 			: lambda : LessThanClause (), 
	"lessThanEquals"	: lambda : LessThanEqualsClause (),
	"greaterThan" 		: lambda : GreaterThanClause (),
	"greaterThanEquals" : lambda : GreaterThanEqualsClause (), 
	"startsWith" 		: lambda : StartsWithClause (),
	"inSet" 			: lambda : InSetClause (),
	"fullText" 			: lambda : FullTextClause (),
	"regExp"			: lambda : RegExpClause (),
	"proximity"			: lambda : ProximityClause (),
	"category"			: lambda : CategoryClause (),
	"and" 				: lambda : AndClause (), 
	"or" 				: lambda : OrClause (),
	"field"				: lambda : Field (),
	"fullTextFields"	: lambda : FullTextFields(),
	"integer"			: lambda : IntegerValue (),
	"string"			: lambda : StringValue (),
	"boolean"			: lambda : BooleanValue (),
	"date"				: lambda : DateValue (),
}

def new_by_name (name, attrs={}):
	obj = CLAUSES[name]()
	obj.set_attributes (attrs)
	return obj


_query_parser_singleton = None

def get_parser ():
	"""
	Static method to get a reference to a singleton parser instance
	"""
	global _query_parser_singleton
	
	if _query_parser_singleton:
		return _query_parser_singleton
	
	_query_parser_singleton = QueryParser()
	return _query_parser_singleton

class QueryParser (xml.sax.ContentHandler):
	def __init__ (self):
		xml.sax.ContentHandler.__init__(self)
		self._query = None
		self._current_parent = None
		self._parents = None # a stack to keep track of parents
	
	def parse (self, query_xml):
		"""
		Return a Query object representing the query described by an
		xml document.
		
		You can pass either a string or file-like object to this method
		"""
		self._parents = []
		
		# xml.sax.parse takes a file-like object, not strings so we have to tweak strings
		if isinstance (query_xml, basestring) :
			xml.sax.parseString(query_xml, self)
		elif isinstance (query_xml, file):	
			xml.sax.parse (query_xml, self)
		else:
			raise IOError ("%s is not a string or file" % query_xml)
		return self._query
		
	def startElement (self, name, attrs):
		if len (self._parents) >= 1:
			parent = self._parents[-1]
		else:
			parent = None
		if name == "request":
			return
			
		child = new_by_name(name, attrs)
		
		if isinstance (child, Query):
			self._query = child
		elif isinstance (child, Clause):
			if isinstance (parent, Query):
				parent.set_clause (child)
			elif isinstance (parent, SelectorClause) :
				raise QueryException ("Trying to append %s to selector %s" % (name,parent.get_name()))
			elif isinstance (parent, CollectorClause) :
				parent.append_clause (child)
			else:
				raise QueryException ("Trying to append %s to unknown parent %s" % (name, parent.get_name()))
		elif isinstance (child, Field):
			if not isinstance (parent, SelectorClause):
				raise QueryException ("Trying to append field to non-selector clause: %s" % parent.get_name())
			child.set_field_name (attrs["name"])
			parent.append_field(child)
		elif isinstance (child, Value):
			if not isinstance (parent, SelectorClause):
				raise QueryException ("Trying to append value %s to non-selector clause: %s" % (child, parent.get_name()))
			if isinstance (parent, MultiValueClause):
				parent.append_value (child)
			else:
				parent.set_value(child)
		elif isinstance (child, FullTextFields):
			if isinstance (parent, ProximityClause) or isinstance (parent, RegExpClause):
				parent.append_field (child)
			else:
				raise QueryException ("Can only append %s to %s or %s" % (child.__class__, ProximityClause, RegExpClause))
		else:
			raise QueryException ("Unknown query element %s" % child)
		
		self._parents.append (child)
		
	def endElement (self, name):
		if len(self._parents) == 0:
			return
		if self._parents[-1].get_name() != name:
			raise QueryException ("Unexpected end tag: " + name + ". Expected " + self._parents[-1].get_name())
		self._parents.pop()
		
	def characters (self, ch):
		if len (self._parents) == 0:
			return
		parent = self._parents[-1]
		if isinstance (parent, Value) :
			parent.set(ch)
		elif isinstance (parent, UserQuery):
			parent.set_string ((parent.get_string() + ch).strip()) # we need to strip newlines
		elif ch.strip() != "" :
			raise QueryException("Element %s does not accept character data: '%s'" % (parent.get_name(), ch))

PHRASE_MODIFIERS = \
{
	"c" : {"caseSensitive": True},
	"C" : {"caseSensitive": False},
	"d" : {"diacriticSensitive": True},
	"D" : {"diacriticSensitive": False},
	"l" : {"enableStemming": False},
	"L" : {"enableStemming": True},
	"e" : {"caseSensitive": True, "diacriticSensitive": True, "enableStemming": False},
	"f" : {"fuzzy": 10},
	"s" : {"slack": 1}, #FIXME a saner default choice depends on number of words in search string
	"w" : {}, # FIXME : Word based matching. There's no parallel in query language!
	"o" : {"ordered": True},
}

CLAUSE_MODIFIERS = \
{

	"b" : {"boost": 10},
	"r" : {},
	"p" : {},
	"-" : {"negate" : True}
}

CLAUSE_QUALIFIERS = \
{
	"r" : "regExp",
	"p" : "proximity"
}

class UserPhrase :
	"""
	Model a phrase string with modifiers, like
	
		"hello world"cpb
	"""
	def __init__ (self, phrase):
		self._phrase = None
		self._modifiers = ""
		self._clause_modifiers = ""
		last_quote = phrase.rfind('"')
		self._negated = (phrase[0] == "-")
		
		if last_quote == 0 or last_quote == -1:
			# There is no terminating quote or no quotes at all
			# just accept the string silently as a phrase
			self._phrase = phrase
		else:
			self._phrase = phrase[1+int(self._negated):last_quote]
			modifiers = phrase[last_quote+1:]
			for mod in modifiers:
				if PHRASE_MODIFIERS.has_key (mod):
					self._modifiers += mod
				elif CLAUSE_MODIFIERS.has_key(mod) or CLAUSE_QUALIFIERS.has_key(mod):
					self._clause_modifiers += mod
				else:
					xesam.error ("Unknown phrase modifier '%s'. Skipping." % mod)
		
		if self._negated:
			self._clause_modifiers += "-"
					
	def is_negated (self):
		"""
		Returns True if there is a - in front of the original string
		"""
		return self._negated
	
	def get_modifiers (self):
		"""
		Return a string containing the modifiers fx 'dcl' if the original
		string was '"foo bar"dcl'. Only modifiers from the original string
		affecting *only* the value to match is returned.
		Modifiers affecting the containing clause can be obtained via
		get_clause_modifiers()
		"""
		return self._modifiers
	
	def get_modifiers_long (self):
		"""
		Return a dict modifierName->Value. This can be used directryl 
		as set_attribute() argument on StringValues for example.
		"""
		result = {}
		for mod in self._modifiers : result.update (PHRASE_MODIFIERS[mod])
		return result
	
	def get_string (self):
		"""
		Get the string contained in the phrase. If the original string
		was '"hello world"dcl' this method returns 'hello world'.
		"""
		return self._phrase
	
	def get_clause_modifiers (self):
		"""
		See get_modifiers(), except this method returns only modifiers
		affecting the parent clause.
		"""
		return self._clause_modifiers
		
	def get_clause_modifiers_long (self):
		"""
		Look at get_clause_modifiers() and get_modifiers_long() and do
		the math :-P
		"""
		result = {}
		for mod in self._clause_modifiers : result.update (CLAUSE_MODIFIERS[mod])
		return result
	
	def string_value (self):
		"""
		Return this phrase as a xesam.query.StringValue
		"""
		return StringValue(self._phrase, attrs=self.get_modifiers_long())


class UserSearchParser :
	def __init__ (self):
		pass
		
	def parse (self, user_string):
		"""
		Return a xesam.query.Clause representing the user search
		string. You can embed this directly into a xesam.query.CompositeQuery
		"""
		terms = self.tokenize(user_string)
		tree = None
		clauses = []
		for i in range(len(terms)):
			clause = None
			if self.is_phrase (terms[i]):
				clause = self.parse_phrase (terms[i])
			elif self.is_select (terms[i]):
				clause = self.parse_select (terms[i])
			elif self.is_collector (terms[i]):
				if i == len(terms) - 1 or i == 0:
					# This is the first or last term, boolean ops ignored
					xesam.error ("Boolean operation '%s' as first or last term in query '%s', parsing as word" % (terms[i], user_string))
					clause = self.parse_word(terms[i])
				else:	
					clause = self.parse_collector(terms[i])
			else:
				clause = self.parse_word (terms[i])
			
			clauses.append(clause)
			
			if isinstance(clause, AndClause):
				if clause.__class__ == tree.__class__:
					pass 
				elif isinstance (tree, OrClause):
					# We need to absorb the last child of previous
					# since AND takes precedence
					children = tree.get_children()
					tree.clear_children()
					tree.append_clauses(children[:-1])
					tree.append_clause (clause)
					clause.append_clause (children[-1])
					tree = clause
				elif not previous is None:
					clause.append_clause (tree)
					tree = clause
			elif isinstance (clause, OrClause):
				if clause.__class__ == tree.__class__:
					pass
				elif not tree is None:
					clause.append_clause (tree)
					tree = clause
				else:
					xesam.error ("parse tree is none")
			else:
				if tree is None:
					tree = clause
				elif isinstance (tree, AndClause):
					tree.append_clause (clause)
				elif isinstance (tree, OrClause)and isinstance(clauses[-2],OrClause):
					tree.append_clause (clause)
				elif isinstance (tree, OrClause):
					collector = AndClause()
					children = tree.get_children()
					tree.clear_children()
					tree.append_clauses(children[:-1])
					collector.append_clause (children[-1])
					collector.append_clause(clause)
					tree.append_clause(collector)
					tree = collector
				else:
					collector = AndClause()
					collector.append_clause (tree)
					collector.append_clause (clause)
					tree = collector
		
		return tree.get_top_parent()
		
		
	def tokenize (self, string):
		tokens = []
		token = ""
		in_phrase = False
		skip = 0
		
		for i in range(len(string)):
			if skip > 0 :
				skip -= 1
				continue
				
			char = string[i]
			if in_phrase:
				if char == '"' :
					# look for modifiers
					while i < len(string) and string[i] != " ":
						token += string[i]
						i += 1
						skip += 1
					tokens.append(token)
					token = ""
					in_phrase = False
					continue
				else:
					token += char
					continue
				
			if char == '"':
				in_phrase = True
				token += char
				continue
			elif char == ':' or char == '=':
				in_select = True
				token += char
				continue
			else:				
				if char == " ":
					tokens.append(token)
					token = ""
				else:
					token += char
				
		if token != "":
			tokens.append(token)
		return tokens
	
	def parse_word (self, word):
		clause = FullTextClause()
		if "" == word:
			raise QueryException("Cannot parse empty word")
		if word[0] == "-":
			clause.set_value (StringValue(word[1:]))
			clause.set_negate(True)
		else:
			clause.set_value(StringValue(word))
		return clause
	
	def is_phrase (self, term):
		rquote = term.rfind ('"')
		if (term[0] == '"' and rquote != 0):
			return True
		elif (term[0]== "-" and term[1] == '"' and rquote != 1):
			return True
		return False
	
	def parse_phrase (self, term):
		phrase = UserPhrase(term)
		clause = None
		for mod in phrase.get_clause_modifiers():
			if mod == "p" or mod == "r":
				clause = CLAUSES[CLAUSE_QUALIFIERS[mod]]()
		
		if clause is None:
			clause = FullTextClause ()
				
		clause.set_attributes (phrase.get_clause_modifiers_long())
		
		if isinstance (clause, ProximityClause):
			for word in phrase.get_string().split(" "):
				clause.append_value (StringValue(word, attrs=phrase.get_modifiers_long()))
			clause.append_field (FullTextFields())
		elif isinstance (clause, RegExpClause):
			clause.set_value (StringValue(phrase.get_string()))
			clause.append_field (FullTextFields())
			if len(phrase.get_modifiers()) > 0:
				xesam.error ("Removed modifiers from regExp " + phrase.get_modifiers())
		elif isinstance (clause, MultiValueClause):
			clause.append_value (phrase.string_value())
		else:
			clause.set_value (phrase.string_value())
		
		return clause
	
	def is_select (self, term):
		return (term.count("=") > 0) or (term.count(":") > 0)
	
	def parse_select (self, term):
		implied_clause = None
		keyword = None
		i = 0
		for i in range(len(term)): 
			if term[i] == "=" :
				implied_clause = EqualsClause()
				break
			elif term[i] == ":":
				implied_clause = ContainsClause()
				break
		keyword = term[:i]
		value = term[i+1:]
		
		# FIXME proper keyword expansion to Fields in user search lang
		fields = [Field("DC:Title")]#self.get_fields_from_keyword (keyword)
		
		if self.is_phrase (value):
			phrase_clause = self.parse_phrase (value)
			
			if isinstance(phrase_clause, ProximityClause) or isinstance (phrase_clause, RegExpClause):
			   	# We should keep this clause type instead of the 
			   	# one implied by : or = relation in selector
			   	for field in fields : phrase_clause.append_field (field)
				return phrase_clause
			else:
				# We should copy value and selector attributes to
				# the clause implied by the : or = relation
				for field in fields : implied_clause.append_field (field)
				implied_clause.set_value (StringValue(value))
				implied_clause.set_attributes(phrase_clause.get_attributes())
				implied_clause.get_value().set_attributes (phrase_clause.get_value().get_attributes())
				return implied_clause
		
		for field in fields : implied_clause.append_field (field)
		implied_clause.set_value (StringValue(value))
		return implied_clause
	
	def is_collector (self, term):
		return    term == "AND" \
			   or term == "&&" \
			   or term == "OR" \
			   or term == "||"
	
	def parse_collector (self, term):
		# This method falls back to an AndClause for unknown collectors
		if term == "OR" or term == "||":
			return OrClause()
		if term != "AND" or term != "&&":
			xesam.debug ("Unknown collector '%s' defaulting to AndClause" % term)
		return AndClause()
