#!/usr/bin/python

# Update python path to use local xesam module
import sys, os
from os.path import dirname, join, abspath
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)+"/.."))

from xesam.query import *
import unittest

class InstantiationTest (unittest.TestCase):
	def testAllInstantiation (self):
		for clause_constructor in xesam.query.CLAUSES.itervalues():
			clause_constructor()

class ParserBootStrap (unittest.TestCase):
	"""
	This class tests that creating a CompositeQuery, serilizing,
	parsing, and re-serializing gives the same serialized query
	again.
	"""
	def testBootStrapSimple (self):
		q1 = CompositeQuery()
		sel = SelectorClause("equals")
		sel.set_value(IntegerValue(1))
		sel.set_boost(20)
		sel.append_field(Field("File:size"))
		sel.append_field(Field("Image:bog"))
		q1.set_clause(sel)
		
		xml1 = q1.xml()
		q2 = QueryParser().parse(xml1)
		xml2 = q2.xml()
		
		self.assertEqual (xml1, xml2)
	
	def testBootStrapComplex (self):
		pass # FIXME
	
	def testBootStrapUserQuery (self):
		q = UserQuery ("Hello World")
		xml1 = q.xml()
		
		xml2 = QueryParser().parse(xml1).xml()
		
		self.assertEqual (xml1, xml2)
	
	def testBootStrapUserQueryWithBadXML (self):
		pass

class SampleTest (unittest.TestCase):
	"""
	Various tests performed on the files in sample/queries
	"""
	def testParseAll (self):
		"""
		Try and parse all queries found is sample/queries
		and calculate their xml representation.
		
		When this is done re-parse the compiled query from its
		own xml representation and verify that the first and
		second pass generate identical xml.
		"""
		sample_dir = join(dirname (__file__), "..", "samples", "queries")
		parser = QueryParser()
		
		for query in os.listdir (sample_dir):
			
			# Skip all but xml files
			if not query.endswith (".xml") : continue 
			
			query = join (sample_dir, query)
			print >> sys.stderr, "Parsing " + query
			parsed_query = parser.parse (file(query))
			parsed_xml = parsed_query.xml()
			
			print >> sys.stderr, "Bootstrapping " + query
			try:
				# FIXME: We need the str() around the parsed_xml
				#        because the sax parser only works on 'str'
				#        and not 'unicode' objects. Duh!
				second_pass = parser.parse (str(parsed_xml))
			except Exception, e:
				self.fail ("Error parsing xml of bootstrapped query: %s\n. %s: %s\n. Processed query:\n%s" % (query, e.__class__, e, parsed_xml))
			second_xml = second_pass.xml() 
						
			self.assertEqual (parsed_xml, second_xml)

class UserSearchPhraseTest (unittest.TestCase):
	"""
	Test various aspects of the xesam.query.UserSearchParser's
	handling of phrases
	"""
	def testRegExpPhrase (self):
		"""
		Try and parse a simple regexp phrase
		"""
		phrase_simple = '"hello|world"r'
		
		parser = UserSearchParser()
		
		clause = parser.parse_phrase (phrase_simple)
		self.assertTrue (isinstance(clause, RegExpClause))
		xesam.debug ("\n"+clause.xml())
	
	def testProximityPhrase (self):
		"""
		Try and parse a simple proximity phrase
		"""
		phrase_simple = '"hello world"p'
		phrase_multi_mod = '"hello world"plc'
		parser = UserSearchParser()
		
		clause = parser.parse_phrase (phrase_simple)
		self.assertTrue (isinstance (clause, ProximityClause))
		xesam.debug ("\n"+clause.xml())
		
		clause = parser.parse_phrase (phrase_multi_mod)
		self.assertTrue (isinstance (clause, ProximityClause))
		xesam.debug ("\n"+clause.xml())
	
	def testIllegalPhraseModifier (self):
		"""
		We shouldn't crash
		"""
		phrase = '"blah di blah"x'
		parser = UserSearchParser()		
		clause = parser.parse_phrase(phrase)
		
		self.assertTrue (isinstance(clause, FullTextClause))
		xesam.debug ("\n"+clause.xml())
	
	def testPhraseMods (self):
		"""
		Test all modifiers from xesam.query.PHRASE_MODIFIERS
		(those are modifiers not changing the default clause type)
		"""
		phrase_simple = '"hello world"'
		parser = UserSearchParser()
		for mod in PHRASE_MODIFIERS:
			clause = parser.parse_phrase (phrase_simple+mod)
			self.assertTrue (isinstance(clause, FullTextClause))
			string = clause.get_value()
			self.assertTrue (string.get_attributes() == PHRASE_MODIFIERS[mod], "For modifier '%s', %s != %s" % (mod, string.get_attributes(), PHRASE_MODIFIERS[mod]))
	
	def testClauseMods (self):
		"""
		Test all modifiers from xesam.query.CLAUSE_MODIFIERS
		(those are modifiers not changing the default clause type)
		"""
		phrase_simple = '"hello world"'
		parser = UserSearchParser()
		for mod in CLAUSE_MODIFIERS:
			clause = parser.parse_phrase (phrase_simple+mod)
			self.assertTrue (clause.get_attributes() == CLAUSE_MODIFIERS[mod], "For modifier '%s', %s != %s" % (mod, clause.get_attributes(), CLAUSE_MODIFIERS[mod]))

class UserPhraseTest (unittest.TestCase):
	def testUserPhrase (self):
		phrase = UserPhrase ('"hello world"brpefsw')
		
		self.assertTrue ("b" in phrase.get_clause_modifiers())
		self.assertTrue ("r" in phrase.get_clause_modifiers())
		self.assertTrue ("p" in phrase.get_clause_modifiers())
		
		self.assertTrue ("e" in phrase.get_modifiers())
		self.assertTrue ("f" in phrase.get_modifiers())
		self.assertTrue ("s" in phrase.get_modifiers())
		self.assertTrue ("w" in phrase.get_modifiers())
		
		self.assertTrue ("hello world" == phrase.get_string())
	
	def testGetLong (self):
		phrase = UserPhrase ('"hello world"brpeflsw')
		mods = phrase.get_modifiers_long ()
		clause_mods = phrase.get_clause_modifiers_long()
		
		self.assertTrue (clause_mods.has_key("boost"))
		self.assertTrue (mods.has_key("caseSensitive"))
		self.assertTrue (mods.has_key("diacriticSensitive"))
		self.assertTrue (mods.has_key("enableStemming"))
		self.assertTrue (mods.has_key("fuzzy"))
		self.assertTrue (mods.has_key("slack"))
		#self.assertTrue (mods.has_key("caseSensitive")) # FIXME word break test
	
	def testNegated (self):
		phrase = UserPhrase ('-"hello world"')
		self.assertTrue (phrase.is_negated())
		

class UserSearchSelectTest (unittest.TestCase):
	"""
	Test various aspects of the xesam.query.UserSearchParser's
	handling of select terms
	"""
	def testContainsSelect (self):
		"""
		Try and parse a foo:bar selector type
		"""
		select_simple = 'foo:bar'
		parser = UserSearchParser()
		
		clause = parser.parse_select (select_simple)
		self.assertTrue (isinstance(clause, ContainsClause))
		xesam.debug ("\n"+clause.xml())
	
	def testEqualsSelect (self):
		select_simple = 'foo=bar'
		parser = UserSearchParser()
		
		clause = parser.parse_select (select_simple)
		self.assertTrue (isinstance(clause, EqualsClause))
		xesam.debug ("\n"+clause.xml())
	
	def testPhraseSelect (self):
		select_eq = 'foo="bar"cd'
		select_con = 'foo:"bar"cd'
		parser = UserSearchParser()
		
		clause_eq = parser.parse_select (select_eq)
		clause_con = parser.parse_select (select_con)
		
		self.assertTrue (isinstance(clause_eq, EqualsClause))
		self.assertTrue (isinstance(clause_con, ContainsClause))
		
		xesam.debug ("\n"+clause_eq.xml())
		xesam.debug ("\n"+clause_con.xml())
	
	def testPhraseSelectWithSelectorMods (self):
		select = 'foo="bar zyg"eb'
		parser = UserSearchParser()
		
		clause = parser.parse_select (select)
		
		self.assertTrue (isinstance(clause, EqualsClause))
		
		# test for a boost attribute
		self.assertTrue (clause.get_attributes()["boost"] != 1.0)
		
		xesam.debug ("\n"+clause.xml())
	
	def testRegExpSelect (self):
		select = 'foo="bar|zyg"r'
		parser = UserSearchParser()
		
		clause = parser.parse_select (select)
		
		self.assertTrue (isinstance(clause, RegExpClause))
		
		xesam.debug ("\n"+clause.xml())
	
	def testProximitySelect (self):
		select = 'foo="bar zyg"p'
		parser = UserSearchParser()
		
		clause = parser.parse_select (select)
		
		self.assertTrue (isinstance(clause, ProximityClause))
		
		xesam.debug ("\n"+clause.xml())

class UserSearchParserTokenizerTest (unittest.TestCase):
	def testTokenizeMultipleWords (self):
		string = "hello how are you"
		correct = ["hello", "how", "are", "you"]
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
		
	def testTokenizeSingleWord (self):
		string = "hello"
		correct = ["hello"]
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeSinglePhrase (self):
		string = '"hello world how are you"'
		correct = ['"hello world how are you"']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeMultiplePhrases (self):
		string = '"hello world" "how are you"'
		correct = ['"hello world"', '"how are you"']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeSinglePhraseWithMods (self):
		string = '"hello world how are you"cdlp'
		correct = ['"hello world how are you"cdlp']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeMultiplePhrasesWithMods (self):
		string = '"hello world"p "how are you"cdl'
		correct = ['"hello world"p', '"how are you"cdl']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeSimpleSelect (self):
		string = 'type:hello'
		correct = ['type:hello']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizePhraseSelect (self):
		string = 'type:"hello world"'
		correct = ['type:"hello world"']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizePhraseWithModsSelect (self):
		string = 'type:"hello world"cp'
		correct = ['type:"hello world"cp']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeMultipleSelects (self):
		string = 'type:image ext=jpg'
		correct = ['type:image', 'ext=jpg']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeMultiplePhrasesWithModsSelect (self):
		string = 'title:"hello world"p ext="jpg|png"r'
		correct = ['title:"hello world"p', 'ext="jpg|png"r']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
	def testTokenizeMixedString (self):
		string = 'type:"image|doc"r "flower africa"p random words'
		correct = ['type:"image|doc"r', '"flower africa"p', 'random', 'words']
		parser = UserSearchParser()
		
		tokens = parser.tokenize(string)
		self.assertTrue (tokens == correct, "%s != %s" % (tokens,correct))
	
if __name__ == '__main__':
	unittest.main()


