#!/usr/bin/python


#
# This is a suite of unit tests designed to exercise all features
# of a xesam search implementation.
#
# It is not the intended purpose of these tests to validate the actual
# search results. Just make sure that they don't throw errors and behave
# as dictated by the xesam search spec.
#


# Update python path to use local xesam module
import sys, os, dbus
from os.path import dirname, join, abspath
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)+"/.."))

from xesam.query import *
from xesam.client import *

import unittest

def get_default_composite_query ():
	query = xesam.query.CompositeQuery()
	selector = xesam.query.FullTextClause()
	selector.set_value(xesam.query.StringValue(xesam.TEST_STRING))
	query.set_clause (selector)
	return query

class SessionMethodInvocationTest (unittest.TestCase):
	"""
	Test that methods related to session managment
	does not throw errors on mere invocation
	"""
	def __init__ (self, arg):
		unittest.TestCase.__init__ (self, arg)
		self.session = None
		
	def setUp (self):
		self.session = Session()
		
	def tearDown (self):
		self.session.close()
	
	def testNewSession (self):
		"""Create a remote session"""
		# this is already tested in the setUp method
		pass
	
	def testCloseSession (self):
		"""Close a remote session"""
		# this is already tested in the tearDown method
		pass
	
	def testGetProperty (self):
		"""Check that we do not crash when trying to get a property"""
		# this is already tested in the tearDown method
		self.session.get_property ("vendor.search.live")
	
	def testSetProperty (self):
		"""Check that we do not crash when trying to set a property"""
		# this is already tested in the tearDown method
		self.session.set_property ("sort.order", ["descending"])

class SessionSanityTest (unittest.TestCase):
	"""
	Test that the methods on the session object behaves (relatively) sane.
	"""
	def __init__ (self, arg):
		unittest.TestCase.__init__ (self, arg)
		self.session = None
		
	def setUp (self):
		self.session = Session()
		
	def tearDown (self):
		self.session.close()
	
	def testSessionCleanup (self):
		"""
		Check that we cannot call methods on a closed session
		"""		
		# We have to bypass the Session object and use dbus directly
		searcher = self.session.get_searcher()
		handle = self.session.get_handle()
		
		self.session.close()				
		try:
			searcher.GetProperty(handle, "search.live")
			self.fail("session.GetProperty must fail after closing the session")
		except:
			pass
	
	def testSessionHandleUniqueness (self):
		"""
		Test that two different session have different handles
		"""
		session2 = Session()
		self.assertTrue (self.session.get_handle() != session2.get_handle(), "session handles should be unique")

class SessionPropertyDefaultsTest (unittest.TestCase):
	"""
	Test that the default session property values are respected
	"""
	def __init__ (self, arg):
		unittest.TestCase.__init__ (self, arg)
		self.session = None
		
	def setUp (self):
		self.session = Session(use_dbus_types=True)
		
	def tearDown (self):
		self.session.close()
	
	def testHitFieldsDefault (self):
		"""
		Assert that the default value for hit.fields is ["xesam:url"]
		"""
		hitfields = self.session.get_property ("hit.fields")
		self.assertTrue (hitfields == ["xesam:url"], "Default hit.fields value must be ['xesam:url'], got %s" % hitfields)
	
	def testHitFieldsExtendedDefault (self):
		"""
		Assert that the default value for hit.fields.extended is [] (an empty list)
		"""
		extf = self.session.get_property ("hit.fields.extended")
		self.assertTrue (extf == [], "Default hit.fields.extended value must be an empty list, found %s" % extf)
		
	def testHitSnippetLengthDefault (self):
		"""
		Assert that the default value for hit.snippet.length is 200
		"""
		self.assertTrue (self.session.get_property ("hit.snippet.length") == 200, "Default hit.snippet.length value must be 200")

	def testSortFieldsDefault (self):
		"""
		Assert that the default value for sort.fields is ["xesam:relevancyRating"]
		"""
		fields = self.session.get_property ("sort.fields")
		self.assertTrue (fields == ["xesam:relevancyRating"], "Default sort.fields value must be ['xesam:relevancyRating'], got '%s'" % fields)

	def testSortOrderDefault (self):
		"""
		Assert that the default value for sort.order is 'descending'
		"""
		order = self.session.do_get_property ("sort.order")
		self.assertTrue (order == [], "Default sort.order value must be an empty list, found '%s'" % order)
	
	def testVendorIdDefault (self):
		"""
		Assert that the default value for vendor.id is a non-empty string
		"""
		vid = self.session.do_get_property ("vendor.id")
		self.assertTrue (vid != None, "Default vendor.id value must not be None")
		self.assertTrue (len(vid) > 0 and isinstance(vid,basestring), "Default vendor.id value must must be a non-empty string")
	
	def testVendorVersionDefault (self):
		"""
		Assert that the default value for vendor.version is a non-empty string
		"""
		version = self.session.do_get_property ("vendor.version")
		self.assertTrue (version != None, "vendor.version must not be None")
		self.assertTrue (type(version) == dbus.UInt32, "vendor.version must be a dbus.UInt32, got '%s' = %s" % (type(version),version))

	def testVendorDisplayDefault (self):
		"""
		Assert that the default value for vendor.display is a string
		"""
		disp = self.session.do_get_property ("vendor.display")
		self.assertTrue (isinstance(disp, basestring) , "vendor.display must be a string, found %s" % type(disp))

	def testVendorXesamDefault (self):
		"""
		Assert that the default value for vendor.xesam is an integer
		"""
		xesam_v = self.session.do_get_property ("vendor.xesam")
		self.assertTrue (type(xesam_v) == dbus.UInt32, "vendor.xesam must be a dbus.UInt32, found %s" % type(xesam_v))
		self.assertTrue (xesam_v == 100, "vendor.xesam should be 100 to match the Xesam 1.0 spec")
	
	def testVendorOntologyFieldsDefault (self):
		"""
		Assert that the default value for vendor.ontology.fields is a list
		"""
		fields = self.session.do_get_property ("vendor.ontology.fields")
		self.assertTrue (type(fields) == dbus.Array, "vendor.ontology.fields must be a list, found %s" % type(fields))
		self.assertTrue (fields.signature == "s", "vendor.ontology.fields must be a list of strings. Got array signature '%s'" % fields.signature)
	
	def testVendorOntologyContentsDefault (self):
		"""
		Assert that the default value for vendor.ontology.contents is a list
		"""
		contents = self.session.do_get_property ("vendor.ontology.contents")
		self.assertTrue (type(contents) == dbus.Array, "vendor.ontology.contents must be a list, found %s" % type(contents))
		self.assertTrue (contents.signature == "s", "vendor.ontology.contents must be a list of strings. Got array signature '%s'" % contents.signature)
	
	def testVendorOntologySourcesDefault (self):
		"""
		Assert that the default value for vendor.ontology.sources is a list
		"""
		sources = self.session.do_get_property ("vendor.ontology.sources")
		self.assertTrue (type(sources) == dbus.Array, "vendor.ontology.sourcess must be a list, found %s" % type(sources))
		self.assertTrue (sources.signature == "s", "vendor.ontology.sources must be a list of strings. Got array signature '%s'" % sources.signature)
	
	def testSearchLiveDefault (self):
		"""
		Assert that the default value for vendor.search.live is True
		"""
		self.assertTrue (self.session.get_property ("vendor.search.live") == True, "Default value for vendor.search.live must be True")
	
	def testVendorExtensionsNamesDefault (self):
		"""
		Assert that the default value for vendor.extensions is a list
		"""
		exts = self.session.do_get_property ("vendor.extensions")
		self.assertTrue (type(exts) == dbus.Array, "vendor.extensions must be a list, found %s" % type(exts))
		self.assertTrue (exts.signature == "s", "vendor.extensions must be a list of strings. Got array signature '%s'" % exts.signature)
			
	def testVendorMaxhits (self):
		"""
		Test that vendor.maxhits is an int/uint and that it is not 0
		"""
		maxhits = self.session.do_get_property ("vendor.maxhits")
		self.assertTrue (type(maxhits) == dbus.UInt32, "vendor.maxhits should be a dbus.UInt32, got '%s'" % type(maxhits))
		self.assertTrue (maxhits > 0, "vendor.maxhits should be > 0")

	def testVendorSortFields (self):
		"""
		Test that vendor.sort.fields is an array of strings and that it is empty
		"""
		sortfields = self.session.do_get_property ("vendor.sort.fields")
		self.assertTrue (type(sortfields) == dbus.Array, "vendor.sort.fields must be a list, found %s" % type(sortfields))
		self.assertTrue (sortfields.signature == "s", "vendor.sort.fields must be a list of strings. Got array signature '%s'" % sortfields.signature)

	def testVendorSortMaxFields (self):
		"""
		Test that vendor.sort.maxfields is an array of strings and that it is empty
		"""
		sortmaxfields = self.session.do_get_property ("vendor.sort.maxfields")
		self.assertTrue (type(sortmaxfields) == dbus.UInt32, "vendor.sort.maxfields must be a dbus.UInt32, found %s" % type(sortmaxfields))
		self.assertTrue (sortmaxfields >= 0, "vendor.sort.maxfields must be a greater than 0. Got %s" % sortmaxfields)		

class SearchSanityTest (unittest.TestCase):	
	"""
	Test that the methods on the search object behaves (relatively) sane.
	"""
	def __init__ (self, arg):
		unittest.TestCase.__init__ (self, arg)
		self.session = None
		
	def setUp (self):
		self.session = Session()
		
	def tearDown (self):
		self.session.close()
		
	def testNewSearchNoStart (self):
		"""Check that we can create a new search and close it immediately"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		search.close()
	
	def testNewSearchWithStart (self):
		"""Check that we can create a new search and close it right after it is started"""
		query = UserQuery (xesam.TEST_STRING)		
		search = Search (self.session, query)
		search.start()
		search.close()	
	
	def testSessionCleanup (self):
		"""
		Create a search and close the session. Now try and start the Search
		directly via dbus (we have to bypass the xesam.client.Session object).
		This should fail since session closing should trigger closing all
		searches too.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		self.session.close()
		
		# We have to bypass the Session object and use dbus directly
		searcher = self.session.get_searcher()
		try:
			searcher.StartSearch(search.get_handle())
			self.fail("StartSearch must fail after closing the parent session")
		except:
			pass
	
	def testGetHitsUnstartedSearch (self):
		"""
		Check that the server raises an error when requesting count from an
		unstarted search.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		try:
			search.get_hits (10)
			self.fail ("get_hits() on unstarted search should raise an error")
		except:
			pass
	
	def testGetHitCountUnstartedSearch (self):
		"""
		Check that the server raises an error when requesting hit count from an
		unstarted search.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		try:
			search.get_hit_count ()
			self.fail ("get_hit_count() on unstarted search should raise an error")
		except:
			pass
	
	def testGetHitDataUnstartedSearch (self):
		"""
		Check that the server raises an error when requesting hit count from an
		unstarted search.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		try:
			search.get_hit_data ([], [])
			self.fail ("get_hit_data() on unstarted search should raise an error")
		except:
			pass
	
	def testGetHitCountReturnType (self):
		"""
		Test count hits. Make sure it returns an int.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		search.start()
		num_hits = search.get_hit_count()
		self.assertTrue (type(num_hits) == int, "get_hit_count should return an int, got %s of type %s" % (num_hits, type(num_hits)))
	
	def testSearchHandleUniqueness (self):
		"""
		Test that to Search instances created from the same query have different
		search handles.
		"""
		query = UserQuery (xesam.TEST_STRING)
		search1 = Search (self.session, query)
		search2 = Search (self.session, query)
		self.assertTrue (search1.get_handle() != search2.get_handle(), "search handles on different searches with same query should be different")
	
	def _wait_for_signal (self, signal, obj, launcher, signal_handler):
		"""
		Generic helper method to wait until a signal is
		emitted or fail if a timeout is reached.
		@param signal_name: a string naming the signal to connect to
		@param obj: the object on which to connect the signal
		@param launcher: a callable invoked before entering the mian loop
		@param signal_handler: if callable will be invoked with signal args
		"""
		loop = gobject.MainLoop()
		timeout = 0
		count = 0
		
		# Local callbacks
		def do_fail ():
			loop.quit()
			self.fail("Timed out waiting for '%s' signal" % signal)
		
		def waiting ():
			print "Waiting for '%s' on %s (10s)" % (signal, obj.__class__)
			return True
		
		def response (*args):
			if callable(signal_handler):
				signal_handler(*args)
			loop.quit()
			
		sd_id = obj.connect (signal, response)
		timeout = gobject.timeout_add (10000, do_fail)
		gobject.timeout_add (1000, waiting)
		
		launcher()
		loop.run()
	
	def _check_done_signal (self, search, hit_count, state, msg):
		"""
		Assert that SearchDone signal params match expectations
		"""
		self.assertTrue(isinstance(search, Search),
		                "search_handle should be a xesam.client.Search, found %s" % type(search))
		
		self.assertTrue(type(hit_count) == long or type(hit_count) == int,
		                "hit_count should be an int, found %s" % type(hit_count))
		self.assertTrue(hit_count >= 0, "hit_count should be >= 0")
		
		self.assertTrue(type(state) == long or type(state) == int,
		                "state should be an int, found %s" % type(state))
		self.assertTrue(state >= 0, "state should be >= 0")
		
		self.assertTrue(type(msg) == str,
		                "msg should be a string, found %s" % type(msg))
	    
	def testEmitsSearchDoneUQ (self):
		"""
		Assert that the SearchDone signal is emitted (timeout 10s),
		when running a user query
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		self._wait_for_signal ("done", search, search.start, self._check_done_signal)
	
	def testEmitsHitsAddedUQ (self):
		"""
		Assert that the HitsAdded signal is emitted (timeout 10s),
		when running a user query
		"""
		query = UserQuery (xesam.TEST_STRING)
		search = Search (self.session, query)
		self._wait_for_signal ("hits-added", search, search.start, None)
	
	def testEmitsSearchDoneCQ (self):
		"""
		Assert that the SearchDone signal is emitted (timeout 10s),
		when running a composite query
		"""
		query = get_default_composite_query()
		search = Search (self.session, query)
		self._wait_for_signal ("done", search, search.start, self._check_done_signal)
	
	def testEmitsHitsAddedCQ (self):
		"""
		Assert that the HitsAdded signal is emitted (timeout 10s),
		when running a composite query.
		"""
		query = get_default_composite_query()
		search = Search (self.session, query)
		self._wait_for_signal ("hits-added", search, search.start, None)
	
	def testGetHitsOnHitsAddedCQ (self):
		"""
		Call GetHits as a reponse to a HitsAdded signal requesting the
		number of hits passed as signal argument.
		"""
		query = get_default_composite_query()
		search = Search (self.session, query)
		
		def hit_getter (search, count):
			hits = search.get_hits(count)
			self.assertEquals (len(hits), count, "Hit count should match the number of hits requested. Expected %s, got %s" % (count, len(hits)))
		
		self._wait_for_signal ("hits-added", search, search.start, hit_getter)
	
	def xtestGetHitsOnHitsAddedUQ (self):
		"""
		FIXME: If we enable this test (remove 'x' prefix), we get wierd errors.
		       Only the UQ or CQ version may be enabled at once...
		Call GetHits as a reponse to a HitsAdded signal requesting the
		number of hits passed as signal argument.
		"""
		query = UserQuery(xesam.TEST_STRING)
		search = Search (self.session, query)
		
		def hit_getter (search, count):
			hits = search.get_hits(count)
			self.assertEquals (len(hits), count, "Hit count should match the number of hits requested. Expected %s, got %s" % (count, len(hits)))
		
		self._wait_for_signal ("hits-added", search, search.start, hit_getter)
	
	def testGetHitsUserQuery (self):
		"""
		Try and retrieve a hit via a query in the USL.
		The xesam.TEST_STRING should return atleast
		one hit since it is in the xesam sources.
		"""
		query = UserQuery (xesam.TEST_STRING)
		self.session.set_property ("hit.fields", ["xesam:url"])
		
		search = Search (self.session, query)
		search.start()
		hits = search.get_hits (1)
		
		self.assertEquals (len(hits), 1, "Exactly one hit should be returned, when one hit is requested")
		
		# The following is also asserted in client.Hit.__init__, but it never hurts to double check
		# We expect to find only "id" and "xesam:url"
		self.assertEquals (hits[0].get_field_count(), 2, "Only the xesam:url field should be the only one returned")
	
	def testGetHitsCompositeQuery (self):
		"""
		Try and retrieve a hit from a composite query.
		The xesam.TEST_STRING should return atleast
		one hit since it is in the xesam sources.
		"""
		query = get_default_composite_query ()
		self.session.set_property ("hit.fields", ["xesam:url"])
		
		search = Search (self.session, query)
		search.start()
		hits = search.get_hits (1)
		
		self.assertEquals (len(hits), 1, "Exactly one hit should be returned, when one hit is requested, got %s" % len(hits))
		
		# The following is also asserted in client.Hit.__init__, but it never hurts to double check
		# We expect to find only "id" and "xesam:url"
		self.assertEquals (hits[0].get_field_count(), 2, "Only the xesam:url field should be the only one returned")

class GetHitDataTests (unittest.TestCase):
	"""
	Tests for GetHitData method
	"""		
	def setUp (self):
		self.session = Session()
		
	def tearDown (self):
		self.session.close()
	
	def testHitIdOffset (self):
		"""
		Check that the url of the first read hit is identical to the url we
		get by requesting the url for the first hit via GetHitData.
		"""
		query = UserQuery (xesam.TEST_STRING)
		self.session.set_property ("hit.fields", ["xesam:url"])
		
		search = Search (self.session, query)
		search.start()
		hits = search.get_hits (1)
		hit = hits[0]
		
		new_hits = search.get_hit_data (hits, ["xesam:url"])
		new_hit = new_hits[0]
		
		self.assertEquals (hit["xesam:url"], new_hit["xesam:url"], "Re-fetching the url of a hit 0 should return original url. Got '%s', expected '%s'." %
							(new_hit["xesam:url"], hit["xesam:url"]))
							
		self.assertEquals (len(new_hits), 1, "Exactly one hit should be returned when data for one hit is requested. Got '%s' hits." % len(new_hits))
	
	def testHitDataLength (self):
		"""
		Assert that the number of return hits and number of requested fields
		match the returned data.
		"""
		query = UserQuery (xesam.TEST_STRING)
		self.session.set_property ("hit.fields", ["xesam:url"])
		
		search = Search (self.session, query)
		search.start()
		hits = search.get_hits (5)
		self.assertEquals (len(hits), 5, "This test requires 5 or more hits to succeed.")
		
		new_hits = search.get_hit_data (hits, ["xesam:url", "xesam:title", "xesam:mimeType"])
		
		self.assertEquals (len(hits), len(new_hits), "When requesting data for '%s' hits exactly '%s' hits should be returned. Got '%s'" %
							(len(hits), len(hits), len(new_hits)))
		
		for hit in new_hits:
			# Remember "id" is also in the hit data so add one to hit field count
			self.assertEquals(hit.get_field_count(), 4, "When requesting 3 fields per hit, exactly three fields should be returned per hit. Got '%s' for hit '%s'" %
								(hit.get_field_count(), hit.get_id()))
				
		
if __name__ == '__main__':
	unittest.main()


